package com.sun.xml.internal.bind.v2.runtime.reflect;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import com.sun.xml.internal.bind.Util;
import com.sun.xml.internal.bind.v2.runtime.reflect.opt.OptimizedAccessorFactory;
import java.util.logging.Level;
import java.lang.reflect.Modifier;
import java.util.logging.Logger;
import java.lang.reflect.Field;
import java.util.HashMap;
import java.util.Arrays;
import java.net.URL;
import java.net.URI;
import java.io.File;
import java.util.Date;
import javax.xml.transform.Source;
import javax.activation.DataHandler;
import java.awt.Image;
import javax.xml.datatype.XMLGregorianCalendar;
import javax.xml.datatype.Duration;
import java.util.Calendar;
import java.lang.reflect.Type;
import com.sun.xml.internal.bind.v2.model.core.Adapter;
import javax.xml.bind.annotation.adapters.XmlAdapter;
import org.xml.sax.SAXException;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Loader;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.UnmarshallingContext;
import com.sun.xml.internal.bind.api.AccessorException;
import com.sun.istack.internal.Nullable;
import com.sun.xml.internal.bind.v2.runtime.JAXBContextImpl;
import java.util.Map;
import javax.xml.bind.JAXBElement;
import java.util.List;
import com.sun.xml.internal.bind.v2.runtime.unmarshaller.Receiver;

public abstract class Accessor<BeanT, ValueT> implements Receiver
{
    public final Class<ValueT> valueType;
    private static List<Class> nonAbstractableClasses;
    private static boolean accessWarned;
    private static final Accessor ERROR;
    public static final Accessor<JAXBElement, Object> JAXB_ELEMENT_VALUE;
    private static final Map<Class, Object> uninitializedValues;
    
    public Class<ValueT> getValueType() {
        return this.valueType;
    }
    
    protected Accessor(Class<ValueT> valueType) {
        this.valueType = valueType;
    }
    
    public Accessor<BeanT, ValueT> optimize(@Nullable JAXBContextImpl context) {
        return this;
    }
    
    public abstract ValueT get(BeanT p0) throws AccessorException;
    
    public abstract void set(BeanT p0, ValueT p1) throws AccessorException;
    
    public Object getUnadapted(BeanT bean) throws AccessorException {
        return this.get(bean);
    }
    
    public boolean isAdapted() {
        return false;
    }
    
    public void setUnadapted(BeanT bean, Object value) throws AccessorException {
        this.set(bean, value);
    }
    
    public void receive(UnmarshallingContext.State state, Object o) throws SAXException {
        try {
            this.set(state.getTarget(), o);
        }
        catch (AccessorException e) {
            Loader.handleGenericException(e, true);
        }
        catch (IllegalAccessError iae) {
            Loader.handleGenericError(iae);
        }
    }
    
    public boolean isValueTypeAbstractable() {
        return !Accessor.nonAbstractableClasses.contains(this.getValueType());
    }
    
    public boolean isAbstractable(Class clazz) {
        return !Accessor.nonAbstractableClasses.contains(clazz);
    }
    
    public final <T> Accessor<BeanT, T> adapt(Class<T> targetType, Class<? extends XmlAdapter<T, ValueT>> adapter) {
        return new AdaptedAccessor(targetType, this, adapter);
    }
    
    public final <T> Accessor<BeanT, T> adapt(Adapter<Type, Class> adapter) {
        return new AdaptedAccessor((Class)Utils.REFLECTION_NAVIGATOR.erasure(adapter.defaultType), this, (Class)adapter.adapterType);
    }
    
    public static <A, B> Accessor<A, B> getErrorInstance() {
        return Accessor.ERROR;
    }
    
    static {
        Accessor.nonAbstractableClasses = Arrays.asList(new Class[] { Object.class, Calendar.class, Duration.class, XMLGregorianCalendar.class, Image.class, DataHandler.class, Source.class, Date.class, File.class, URI.class, URL.class, Class.class, String.class, Source.class });
        Accessor.accessWarned = false;
        ERROR = new Accessor<Object, Object>(Object.class) {
            @Override
            public Object get(Object o) {
                return null;
            }
            
            @Override
            public void set(Object o, Object o1) {
            }
        };
        JAXB_ELEMENT_VALUE = new Accessor<JAXBElement, Object>(Object.class) {
            @Override
            public Object get(JAXBElement jaxbElement) {
                return jaxbElement.getValue();
            }
            
            @Override
            public void set(JAXBElement jaxbElement, Object o) {
                jaxbElement.setValue(o);
            }
        };
        (uninitializedValues = new HashMap()).put(Byte.TYPE, Byte.valueOf((byte)0));
        Accessor.uninitializedValues.put(Boolean.TYPE, Boolean.valueOf(false));
        Accessor.uninitializedValues.put(Character.TYPE, Character.valueOf('\0'));
        Accessor.uninitializedValues.put(Float.TYPE, Float.valueOf(0.0f));
        Accessor.uninitializedValues.put(Double.TYPE, Double.valueOf(0.0));
        Accessor.uninitializedValues.put(Integer.TYPE, Integer.valueOf(0));
        Accessor.uninitializedValues.put(Long.TYPE, Long.valueOf(0L));
        Accessor.uninitializedValues.put(Short.TYPE, Short.valueOf((short)0));
    }
    
    public static class FieldReflection<BeanT, ValueT> extends Accessor<BeanT, ValueT>
    {
        public final Field f;
        private static final Logger logger;
        
        public FieldReflection(Field f) {
            this(f, false);
        }
        
        public FieldReflection(Field f, boolean supressAccessorWarnings) {
            super(f.getType());
            this.f = f;
            int mod = f.getModifiers();
            if (Modifier.isPublic(mod) && !Modifier.isFinal(mod) && Modifier.isPublic(f.getDeclaringClass().getModifiers())) {
                return;
            }
            try {
                f.setAccessible(true);
            }
            catch (SecurityException e) {
                if (!Accessor.accessWarned && !supressAccessorWarnings) {
                    Accessor.FieldReflection.logger.log(Level.WARNING, Messages.UNABLE_TO_ACCESS_NON_PUBLIC_FIELD.format(new Object[] { f.getDeclaringClass().getName(), f.getName() }), e);
                }
                Accessor.accessWarned = true;
            }
        }
        
        @Override
        public ValueT get(BeanT bean) {
            try {
                return this.f.get(bean);
            }
            catch (IllegalAccessException e) {
                throw new IllegalAccessError(e.getMessage());
            }
        }
        
        @Override
        public void set(BeanT bean, ValueT value) {
            try {
                if (value == null) {
                    value = Accessor.uninitializedValues.get(this.valueType);
                }
                this.f.set(bean, value);
            }
            catch (IllegalAccessException e) {
                throw new IllegalAccessError(e.getMessage());
            }
        }
        
        @Override
        public Accessor<BeanT, ValueT> optimize(JAXBContextImpl context) {
            if (context != null && context.fastBoot) {
                return this;
            }
            Accessor<BeanT, ValueT> acc = OptimizedAccessorFactory.get(this.f);
            if (acc != null) {
                return acc;
            }
            return this;
        }
        
        static {
            logger = Util.getClassLogger();
        }
    }
    
    public static final class ReadOnlyFieldReflection<BeanT, ValueT> extends Accessor.FieldReflection<BeanT, ValueT>
    {
        public ReadOnlyFieldReflection(Field f, boolean supressAccessorWarnings) {
            super(f, supressAccessorWarnings);
        }
        
        public ReadOnlyFieldReflection(Field f) {
            super(f);
        }
        
        @Override
        public void set(BeanT bean, ValueT value) {
        }
        
        @Override
        public Accessor<BeanT, ValueT> optimize(JAXBContextImpl context) {
            return this;
        }
    }
    
    public static class GetterSetterReflection<BeanT, ValueT> extends Accessor<BeanT, ValueT>
    {
        public final Method getter;
        public final Method setter;
        private static final Logger logger;
        
        public GetterSetterReflection(Method getter, Method setter) {
            super((getter != null) ? getter.getReturnType() : setter.getParameterTypes()[0]);
            this.getter = getter;
            this.setter = setter;
            if (getter != null) {
                this.makeAccessible(getter);
            }
            if (setter != null) {
                this.makeAccessible(setter);
            }
        }
        
        private void makeAccessible(Method m) {
            if (Modifier.isPublic(m.getModifiers()) && Modifier.isPublic(m.getDeclaringClass().getModifiers())) {
                return;
            }
            try {
                m.setAccessible(true);
            }
            catch (SecurityException e) {
                if (!Accessor.accessWarned) {
                    Accessor.GetterSetterReflection.logger.log(Level.WARNING, Messages.UNABLE_TO_ACCESS_NON_PUBLIC_FIELD.format(new Object[] { m.getDeclaringClass().getName(), m.getName() }), e);
                }
                Accessor.accessWarned = true;
            }
        }
        
        @Override
        public ValueT get(BeanT bean) throws AccessorException {
            try {
                return this.getter.invoke(bean, new Object[0]);
            }
            catch (IllegalAccessException e) {
                throw new IllegalAccessError(e.getMessage());
            }
            catch (InvocationTargetException e2) {
                throw this.handleInvocationTargetException(e2);
            }
        }
        
        @Override
        public void set(BeanT bean, ValueT value) throws AccessorException {
            try {
                if (value == null) {
                    value = Accessor.uninitializedValues.get(this.valueType);
                }
                this.setter.invoke(bean, new Object[] { value });
            }
            catch (IllegalAccessException e) {
                throw new IllegalAccessError(e.getMessage());
            }
            catch (InvocationTargetException e2) {
                throw this.handleInvocationTargetException(e2);
            }
        }
        
        private AccessorException handleInvocationTargetException(InvocationTargetException e) {
            Throwable t = e.getTargetException();
            if ((t instanceof RuntimeException)) {
                throw (RuntimeException)t;
            }
            if ((t instanceof Error)) {
                throw (Error)t;
            }
            return new AccessorException(t);
        }
        
        @Override
        public Accessor<BeanT, ValueT> optimize(JAXBContextImpl context) {
            if (this.getter == null || this.setter == null) {
                return this;
            }
            if (context != null && context.fastBoot) {
                return this;
            }
            Accessor<BeanT, ValueT> acc = OptimizedAccessorFactory.get(this.getter, this.setter);
            if (acc != null) {
                return acc;
            }
            return this;
        }
        
        static {
            logger = Util.getClassLogger();
        }
    }
    
    public static class GetterOnlyReflection<BeanT, ValueT> extends Accessor.GetterSetterReflection<BeanT, ValueT>
    {
        public GetterOnlyReflection(Method getter) {
            super(getter, null);
        }
        
        @Override
        public void set(BeanT bean, ValueT value) throws AccessorException {
            throw new AccessorException(Messages.NO_SETTER.format(new Object[] { this.getter.toString() }));
        }
    }
    
    public static class SetterOnlyReflection<BeanT, ValueT> extends Accessor.GetterSetterReflection<BeanT, ValueT>
    {
        public SetterOnlyReflection(Method setter) {
            super(null, setter);
        }
        
        @Override
        public ValueT get(BeanT bean) throws AccessorException {
            throw new AccessorException(Messages.NO_GETTER.format(new Object[] { this.setter.toString() }));
        }
    }
}
